package bin;

import shared.Tools;
import structures.ByteBuilder;
import tax.TaxNode;
import tax.TaxTree;
import tracker.EntropyTracker;

public class Clade extends BinObject implements Comparable<Clade>{

	public Clade(int taxID_, int level_, String name_) {
		taxID=taxID_;
		level=level_;
		name=name_;
		counts=new long[6][];
		counts[1]=new long[5];
		counts[2]=new long[16];
		counts[3]=new long[canonicalKmers[3]];
		counts[4]=new long[canonicalKmers[4]];
		counts[5]=new long[canonicalKmers[5]];
	}

	public static Clade makeClade(int tid) {
		TaxNode tn=tree.getNode(tid);
		assert(tn!=null);
		if(tn==null) {
			return new Clade(tid, -1, null);
		}
		return new Clade(tn.id, tn.level, tn.name);
	}
	
	public synchronized void add(byte[] seq, EntropyTracker et) {
		finished=false;
		countKmersMulti(seq, counts, 5);
		float seqEntropy=et.averageEntropy(seq, false);
		entropy=(entropy*bases+seqEntropy*seq.length)/(float)(bases+seq.length);
		
		bases+=seq.length;
		contigs++;
	}
	
	public synchronized void add(Clade c) {
		finished=false;
		synchronized(c) {
			assert(c.taxID>0) : "\n"+this+"\n"+c+"\n";
			assert(c.bases>0) : "\n"+this+"\n"+c+"\n";//Not really necessary, but preventable
			assert(taxID==c.taxID || (taxID<1 && bases==0)) : "\n"+this+"\n"+c+"\n";
			if(taxID!=c.taxID) {
				assert(taxID<0 && bases==0);
				taxID=c.taxID;
				level=c.level;
				name=c.name;
			}
			
			Tools.add(counts, c.counts);
			entropy=(entropy*bases+c.entropy*c.bases)/(float)(bases+c.bases);

			bases+=c.bases;
			contigs+=c.contigs;
		}
	}
	
	public synchronized void finish() {
		if(finished) {return;}
		gc=calcGC();
		strandedness=EntropyTracker.strandedness(counts[2], 2);
		gcCompEntropy=AdjustEntropy.compensate(gc, entropy);
		finished=true;
	}
	
	public synchronized float calcGC() {
		long[] acgtn=counts[1];
		long a=acgtn[0], c=acgtn[1], g=acgtn[2], t=acgtn[3];
		return (float)((g+c)/Math.max(a+g+c+t, 1.0));
	}
	
	public synchronized void clear() {
		finished=false;
		
		taxID=level=-1;
		name=null;
		
		bases=contigs=0;
		gc=entropy=gcCompEntropy=strandedness=0;
		Tools.fill(counts, 0);
	}
	
	@Override
	public int compareTo(Clade b) {
		if(gc!=b.gc) {return gc>b.gc ? 1 : -1;}
		if(bases!=b.bases) {return bases>b.bases ? 1 : -1;}
		return taxID-b.taxID;
	}
	
	public synchronized String toString() {
		ByteBuilder bb=new ByteBuilder();
		bb.append("tid=").append(taxID).append("\tgc=").append(gc, 4).append("\tname=").append(name);
		return bb.toString();
	}
	
	public synchronized ByteBuilder toBytes(ByteBuilder bb) {
		if(bb==null) {bb=new ByteBuilder();}
		assert(finished);
		if(!finished) {finish();}
		bb.append("tid\t").append(taxID).nl();
		bb.append("level\t").append(level).tab().append(TaxTree.levelToString(level)).nl();
		bb.append("name\t").append(name).nl();
		bb.append("gc\t").append(gc, 4).nl();
		bb.append("entropy\t").append(entropy, 8).nl();
		bb.append("strandedness\t").append(strandedness, 8).nl();
		bb.append("bases\t").append(bases).nl();
		bb.append("contigs\t").append(contigs).nl();
		for(int k=1; k<counts.length; k++) {
			if(counts[k]!=null) {
				bb.append(k).append("mers\t").append(counts[k], '\t').nl();
			}
		}
		return bb;
	}
	
	public synchronized boolean finished() {return finished;}
	
	public int taxID=-1;
	public int level=-1;
	public String name=null;
	public final long[][] counts;
	
	public long bases;
	public long contigs;
	public float gc;
	public float entropy;
	public float gcCompEntropy;
	public float strandedness;
	private boolean finished=false;
	
}
