package bin;

import structures.ByteBuilder;
import tax.TaxTree;

public class Comparison implements Comparable<Comparison> {
	
	public Comparison() {}
	
	public Comparison(Clade query_, Clade ref_) {
		compare(query_, ref_, 1, 1);
	}
	
	float compare(Clade query_, Clade ref_, float maxGC, float maxK5) {
		query=query_;
		ref=ref_;
		assert(query.finished());
		assert(ref.finished());
		clearDif();
		
		gcdif=Math.abs(query.gc-ref.gc);
		if(gcdif>maxGC) {return 1+gcdif;};
		entdif=Math.abs(query.gcCompEntropy-ref.gcCompEntropy);
		strdif=Math.abs(query.strandedness-ref.strandedness);
		if(method==ABS) {return compareABS(maxK5);}
		if(method==COS) {return compareCOS(maxK5);}
		if(method==HEL) {return compareHEL(maxK5);}
		if(method==EUC) {return compareEUC(maxK5);}
		assert(false) : "Bad method: "+method;
		return 1;
	}
	
	float compareABS(float maxK5) {
		k3dif=SimilarityMeasures.absDif(query.counts[3], ref.counts[3]);
		if(k3dif*comparisonCutoffMult2>maxK5) {return k3dif*comparisonCutoffMult2;}
		k4dif=SimilarityMeasures.absDif(query.counts[4], ref.counts[4]);
		if(k4dif*comparisonCutoffMult>maxK5) {return k4dif*comparisonCutoffMult;}
		k5dif=SimilarityMeasures.absDif(query.counts[5], ref.counts[5]);
		return k5dif;
	}
	
	float compareCOS(float maxK5) {
		k3dif=SimilarityMeasures.cosineDifference(query.counts[3], ref.counts[3]);
		if(k3dif*comparisonCutoffMult2>maxK5) {return k3dif*comparisonCutoffMult2;}
		k4dif=SimilarityMeasures.cosineDifference(query.counts[4], ref.counts[4]);
		if(k4dif*comparisonCutoffMult>maxK5) {return k4dif*comparisonCutoffMult;}
		k5dif=SimilarityMeasures.cosineDifference(query.counts[5], ref.counts[5]);
		return k5dif;
	}
	
	float compareHEL(float maxK5) {
		k3dif=SimilarityMeasures.hellingerDistance(query.counts[3], ref.counts[3]);
		if(k3dif*comparisonCutoffMult2>maxK5) {return k3dif*comparisonCutoffMult2;}
		k4dif=SimilarityMeasures.hellingerDistance(query.counts[4], ref.counts[4]);
		if(k4dif*comparisonCutoffMult>maxK5) {return k4dif*comparisonCutoffMult;}
		k5dif=SimilarityMeasures.hellingerDistance(query.counts[5], ref.counts[5]);
		return k5dif;
	}
	
	float compareEUC(float maxK5) {
		k3dif=SimilarityMeasures.euclideanDistance(query.counts[3], ref.counts[3]);
		if(k3dif*comparisonCutoffMult2>maxK5) {return k3dif*comparisonCutoffMult2;}
		k4dif=SimilarityMeasures.euclideanDistance(query.counts[4], ref.counts[4]);
		if(k4dif*comparisonCutoffMult>maxK5) {return k4dif*comparisonCutoffMult;}
		k5dif=SimilarityMeasures.euclideanDistance(query.counts[5], ref.counts[5]);
		return k5dif;
	}
	
	void clearDif() {
		gcdif=entdif=strdif=k3dif=k4dif=k5dif=1;
	}
	
	public void setFrom(Comparison b) {
		query=b.query;
		ref=b.ref;
		gcdif=b.gcdif;
		entdif=b.entdif;
		strdif=b.strdif;
		k3dif=b.k3dif;
		k4dif=b.k4dif;
		k5dif=b.k5dif;
	}
	
	public String toString() {
		if(ref==null) {return null;}
		ByteBuilder bb=new ByteBuilder();
		bb.append("tid=").append(ref.taxID).append("\tname=").append(ref.name).nl();
		bb.append("gcdif=").append(gcdif, 5).append("\tedif=").append(entdif, 5);
		bb.append("\tsdif=").append(strdif, 5);
		bb.append("\tk3dif=").append(k3dif, 6).append("\tk4dif=").append(k4dif, 6);
		bb.append("\tk5dif=").append(k5dif, 6);
		return bb.toString();
	}
	
	@Override
	public int compareTo(Comparison b) {
		if(k5dif!=b.k5dif) {return k5dif<b.k5dif ? -1 : 1;}
		if(k4dif!=b.k4dif) {return k4dif<b.k4dif ? -1 : 1;}
		if(k3dif!=b.k3dif) {return k3dif<b.k3dif ? -1 : 1;}
		if(gcdif!=b.gcdif) {return gcdif<b.gcdif ? -1 : 1;}
		if(entdif!=b.entdif) {return entdif<b.entdif ? -1 : 1;}
		if(strdif!=b.strdif) {return strdif<b.strdif ? -1 : 1;}
		return 0;
	}
	
	boolean correct() {
		return ref!=null && query.taxID!=0 && query.taxID==ref.taxID;
	}
	
	boolean incorrect() {
		return ref!=null && query.taxID!=0 && ref.taxID!=0 && query.taxID!=ref.taxID;
	}
	
	int correctLevel() {
		if(ref==null || query==null || query.taxID==0 || 
				ref.taxID==0 || BinObject.tree==null) {
			return TaxTree.LIFE;
		}
		return BinObject.tree.commonAncestorLevel(query.taxID, ref.taxID);
	}
	
	public static void setComparisonCutoffMult(float f) {
		assert(f>=1);
		comparisonCutoffMult=f;
		comparisonCutoffMult=f*f;
	}
	
	Clade query, ref;
	
	float gcdif=1;
	float entdif=1;
	float strdif=1;
	float k3dif=1;
	float k4dif=1;
	float k5dif=1;

	private static float comparisonCutoffMult=2f;
	private static float comparisonCutoffMult2=4f;
	static final int ABS=1, COS=2, HEL=3, EUC=4;
	static int method=ABS;
	
}
